from collections import OrderedDict
from typing import Iterable, Tuple, Union

import numpy as np
import torch
import torch.nn as nn

def create_slice_from_str(indices: str) -> slice:
    """Creates a slice object from a string representing the slice.

    :param indices: String representing the slice, e.g. ``...,3:5,:``
    :type indices: str
    :return: Slice object corresponding to the input indices.
    :rtype: slice
    """
    if len(indices) == 0:
        raise ValueError('Empty string is not a valid slice.')
    return eval(f'np.s_[{indices}]')


def recursive_to_device(
    iterable: Union[Iterable, torch.Tensor], 
    device: Union[str, torch.device],
) -> Iterable:
    """Recursively puts an Iterable of (Iterable of (...)) tensors on the given device

    :param iterable: Tensor or Iterable of tensors or iterables of ...
    :type iterable: Tensor or Iterable
    :param device: Device on which to put the object
    :type device: Union[str, torch.device]
    :raises TypeError: Unexpected tyes
    :return: Nested iterable with the tensors on the new device
    :rtype: Iterable
    """
    if isinstance(iterable, torch.Tensor):
        return iterable.to(device)

    new = []
    for i, item in enumerate(iterable):
        if isinstance(item, torch.Tensor):
            new.append(item.to(device))
        elif isinstance(item, Iterable):
            new.append(recursive_to_device(item, device))
        else:
            raise TypeError(f'Expected type tensor or Iterable but got {type(item)}.')
    if isinstance(iterable, Tuple):
        new = tuple(new)
    return new


#TODO for version 1.0.0: Remove this file
import logging

from .common_hooks import additive_output_noise, hidden_patch_hook_fn

logging.warning("rome_hooks.py is deprecated and will be removed in version 1.0.0. Please use common_hooks.py instead.")




# Functions that are usually only called once during setup of the interface
import importlib
import json
from typing import Optional, List, Tuple

import streamlit as st
import torch

from .commons import HF_MODELS
from ...transformers_util import get_num_layers, load_from_pretrained
from ...hooks import HookedModel

def create_model_config(model_names):
    with st.form('model_config'):
        st.write('## Model Config')

        if model_names is None:
            model_options = list()
        else:
            model_options = model_names
        
        st.selectbox(
            'Model', 
            options=model_options,
            key='model_name',
            index=0,
        )

        devices = ['cpu']
        if torch.cuda.is_available():
            devices += ['cuda']
        st.selectbox(
            'Device',
            options=devices,
            index=0,
            key='device'
        )
        
        st.text_area(label='Prefix Prompt', key='prefix_prompt', value='')
            
        submitted = st.form_submit_button("Save model config")
        if submitted:
            st.session_state.model, st.session_state.tokenizer, st.session_state.config = on_config_submit(st.session_state.model_name)
            st.write('Model config saved!')


def create_sidebar():
    st.checkbox('Show only local models', value=False, key='local_only')
    
    if not st.session_state.local_only:
        model_names = st.session_state.registered_model_names + HF_MODELS
    else:
        model_names = st.session_state.registered_model_names
    
    create_model_config(model_names)

def create_sample_sliders():
    st.slider(label="Temperature", min_value=0., max_value=1.0, value=0., step=0.01, key='temperature', on_change=on_sampling_config_change)
    st.slider(label="Response length", min_value=1, max_value=1024, value=64, step=1, key='response_length', on_change=on_sampling_config_change)
    st.slider(label="Top P", min_value=0., max_value=1.0, value=1., step=0.01, key='top_p', on_change=on_sampling_config_change)
    st.slider(label="Repetition Penalty (1 = no penalty)", min_value=0.01, max_value=1.0, value=1., step=0.01, key='repetition_penalty', on_change=on_sampling_config_change)
    st.slider(label="Number of Beams", min_value=1, max_value=10, value=1, step=1, key='num_beams', on_change=on_sampling_config_change)

def load_registered_models(model_file_path: str = './registered_models.json') -> None:
    try:
        with open(model_file_path, 'r') as f:
            st.session_state.registered_models = json.load(f)
    except FileNotFoundError:
        st.warning(f"Did not find a 'registered_models.json'. Only showing HF models")
        st.session_state.registered_models = dict()
    st.session_state.registered_model_names = list(st.session_state.registered_models.keys())


def startup(variables: List[str], mode_file_path: Optional[str] = './registered_models.json') -> None:
    """Performs startup tasks for the app.

    :param variables: List of variable names that should be intialized.
    :type variables: List[str]
    :param model_file_path: Path to the file containing the registered models.
    :type model_file_path: Optional[str]
    """
    
    if 'startup_done' not in st.session_state:    
        # set wide layout
        st.set_page_config(layout="wide")
        
        # initialize session state variables
        init_session_state(variables)
        st.session_state['visualization'] = dict()
        st.session_state['startup_done'] = True
        
        # load externally registered models
        load_registered_models(mode_file_path)

def init_session_state(variables: List[str]) -> None:
    """Initialize session state variables to None.

    :param variables: List of variable names to initialize.
    :type variables: List[str]
    """
    for var in variables:
        if var not in st.session_state:
            st.session_state[var] = None
            
def on_config_submit(model_name: str) -> Tuple:
    """Function that is called on submitting the config form.

    :param model_name: Name of the model that should be loaded
    :type model_name: str
    :return: Model, tokenizer, config
    :rtype: Tuple
    """
    # load model, hook it and put it on device and in eval mode
    model, tokenizer, config = load_model(model_name)
    model = HookedModel(model)
    model.to(st.session_state.device).eval()
    
    if model_name in st.session_state.registered_models:
        st.session_state.num_layers = get_num_layers(model, config['layer_key_prefix'])
    else:
        st.session_state.num_layers = get_num_layers(model, 'transformer->h')

    return model, tokenizer, config

def on_sampling_config_change():
    st.session_state.sample_kwargs = dict(
        temperature=st.session_state.temperature,
        max_length=st.session_state.response_length,
        top_p=st.session_state.top_p,
        repetition_penalty=1/st.session_state.repetition_penalty,
        num_beams=st.session_state.num_beams
    )

@st.experimental_singleton
def load_model(model_name: str) -> Tuple:
    """Load the specified model with its tokenizer and config.

    :param model_name: Model name, e.g. 'gpt2-xl'
    :type model_name: str
    :return: Model, Tokenizer, Config
    :rtype: Tuple
    """
    if model_name in st.session_state.registered_model_names:
        # import model constructor
        constructor = st.session_state.registered_models[model_name]['constructor']
        constructor_module = importlib.import_module('.'.join(constructor.split('.')[:-1]))
        constructor_class = getattr(constructor_module, constructor.split('.')[-1])
        
        # load model from checkpoint --> make sure that your class has this method, it's default for pl.LightningModules
        checkpoint = st.session_state.registered_models[model_name]['checkpoint'] 
        model = constructor_class.load_from_checkpoint(checkpoint)

        # load tokenizer
        tokenizer = st.session_state.registered_models[model_name]['tokenizer'] # TODO how to deal with this?
        tokenizer_module = importlib.import_module('.'.join(tokenizer.split('.')[:-1]))
        tokenizer_class = getattr(tokenizer_module, tokenizer.split('.')[-1])
        tokenizer = tokenizer_class()

        # load config
        config = st.session_state.registered_models[model_name]['config']

    else: # attempt to load from huggingface
        model, tokenizer, config = load_from_pretrained(model_name)

    return model, tokenizer, config
    


# define some global variables
HF_MODELS = [
    'gpt2', 
    'gpt2-medium', 
    'gpt2-large', 
    'gpt2-xl',
    'gpt-neo-125M',
    'gpt-neo-1.3B',
    'gpt-neo-2.7B',
    'gpt-j-6b',
]

SESSION_STATE_VARIABLES = [
    'model',
    'tokenizer',
    'config',
    'registered_models',
    'registered_model_names',
    'num_layers',
]



import logging
from typing import Optional, List

import torch
import torch.nn.functional as F
from transformers import AutoTokenizer

from .transformers_util import get_num_layers
from .hooks.common_hooks import create_logit_hook
from .hooks.commons import HookedModel

def generate_logit_lense(
    model: HookedModel, 
    tokenizer: AutoTokenizer, 
    sentence: str,
    layers: Optional[List[int]] = None,
    ranks: Optional[bool] = False,
    kl_div: Optional[bool] = False,
    include_input: Optional[bool] = False,
    layer_key_prefix: Optional[str] = None,
):
    """Generates the necessary data to generate the plots from the logits `lense post 
    <https://www.lesswrong.com/posts/AcKRB8wDpdaN6v6ru/interpreting-gpt-the-logit-lens>`_.

    Returns None for ranks and kl_div if not specified.

    :param model: Model that is investigated.
    :type model: HookedModel
    :param tokenizer: Tokenizer of the model.
    :type tokenizer: AutoTokenizer
    :param sentence: Sentence to be analyzed.
    :type sentence: str
    :param layers: List of layers to be investigated.
    :type layers: Optional[List[int]]
    :param ranks: Whether to return ranks of the correct token throughout layers, defaults to False
    :type ranks: Optional[bool], optional
    :param kl_div: Whether to return the KL divergence between intermediate probabilities and final output probabilities, defaults to False
    :type kl_div: Optional[bool], optional
    :param include_input: Whether to include the immediate logits/ranks/kld after embedding the input, defaults to False
    :type include_input: Optional[bool], optional
    :param layer_key_prefix: Prefix for the layer keys, e.g. 'transformer->h' for GPT like models, defaults to None
    :type layer_key_prefix: Optional[str], optional
    :return: logits, ranks, kl_div
    :rtype: Tuple[torch.Tensor]
    """
    
    # TODO
    if include_input:
        logging.warning("include_input is not implemented yet")
    
    # prepare model input
    tokenized_sentence = tokenizer.encode(sentence, return_tensors='pt').to(model.device)
    targets = tokenizer.encode(sentence)[1:]
    
    # instantiate hooks
    num_layers = get_num_layers(model, layer_key_prefix=layer_key_prefix)
    if layers is None:
        layers = list(range(num_layers))
    logit_hooks = [create_logit_hook(layer, model, 'lm_head', layer_key_prefix) for layer in layers]
    
    # run model
    model.forward(tokenized_sentence, hooks=logit_hooks)
    logits = torch.stack([model.save_ctx[str(layer) + '_logits']['logits'] for layer in range(num_layers)], dim=0)
    logits = F.log_softmax(logits, dim=-1)
    
    # compute ranks and kld
    if ranks:
        inverted_ranks = torch.argsort(logits, dim=-1, descending=True)
        ranks = torch.argsort(inverted_ranks, dim=-1) + 1
        ranks = ranks[:, torch.arange(len(targets)), targets]
    else:
        ranks = None

    if kl_div: # Note: logits are already normalized internally by the logit_hook
        kl_div = F.kl_div(logits, logits[-1][None], reduction='none', log_target=True).sum(dim=-1)
        kl_div = kl_div[:, torch.arange(len(targets)), targets]
    else:
        kl_div = None    
        
    logits = logits[:, torch.arange(len(targets)), targets]
    
    return logits, ranks, kl_div





import json
import argparse
import os

def main(args):
    with open(args.file, 'r') as f:
        data = json.load(f)

    # get model_name from path
    filename = os.path.basename(args.file)
    model_name = filename.split('.')[0]
    print(f"model_name: {model_name}")
    for i in range(2):
        if f'col_{i}' in data:
            for j in range(len(data[f'col_{i}'])):
                os.makedirs(f'{args.target_folder}/col_{i}', exist_ok=True)
                with open(f'{args.target_folder}/col_{i}/layer_{j}.txt', 'w') as f:
                    json.dump(data[f'col_{i}'][f'layer_{j}'], f)
        else:
            print(f"col_{i} not in data -> skipping")

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--file', type=str, required=True)
    parser.add_argument('--target_folder', type=str, required=True)
    args = parser.parse_args()
    
    main(args)



# pre-implemented common hooks
import logging
import math
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import einops
import torch
import torch.nn.functional as F
from tqdm import tqdm

from . import util
from .commons import Hook, HookedModel


def save_output(cpu: bool = True, detach: bool = True) -> Callable:
    """Basic hooking function for saving the output of a module to the global context object

    :param cpu: Whether to save the output to cpu.
    :type cpu: bool
    :param detach: Whether to detach the output.
    :type detach: bool
    :return: Function that saves the output to the context object.
    :rtype: Callable
    """
    def inner(save_ctx, input, output):
        if detach:
            if isinstance(output, torch.Tensor):
                output = output.detach()
            else:
                logging.warning('Detaching tensor for iterables not implemented')
                # not implemented
        if cpu:
            if isinstance(output, torch.Tensor):
                output = output.cpu()
            elif isinstance(output, Iterable): # hope for the best
                output = util.recursive_to_device(output, 'cpu')
            else:
                raise TypeError(f"Unsupported type for output {type(output)}")
        save_ctx['output'] = output

    return inner

def replace_activation(indices: str, replacement_tensor: torch.Tensor, tuple_index: int = None) -> Callable:
    """Creates a hook which replaces a module's activation (output) with a replacement tensor. 
    If there is a dimension mismatch, the replacement tensor is copied along the leading dimensions of the output.

    Example: If the activation has shape ``(B, T, D)`` and replacement tensor has shape ``(D,)`` which you want to plug in
    at position t in the T dimension for every tensor in the batch, then indices should be ``:,t,:``. 

    :param indices: Indices at which to insert the replacement tensor
    :type indices: str
    :param replacement_tensor: Tensor that is filled in.
    :type replacement_tensor: torch.Tensor
    :param tuple_index: Index of the tuple in the output of the module.
    :type tuple_index: int
    :return: Function that replaces part of a given tensor with replacement_tensor
    :rtype: Callable
    """
    slice_ = util.create_slice_from_str(indices)
    def func(save_ctx, input, output):
        if tuple_index is None:
            # add dummy dimensions if shape mismatch
            diff = len(output[slice_].shape) - len(replacement_tensor.shape)
            rep = replacement_tensor[(None,)*diff].to(input.device)
            # replace part of tensor
            output[slice_] = rep
        else:
            # add dummy dimensions if shape mismatch
            diff = len(output[tuple_index][slice_].shape) - len(replacement_tensor.shape)
            rep = replacement_tensor[(None,)*diff].to(input.device)
            # replace part of tensor
            output[tuple_index][slice_] = rep
        return output

    return func

def transformers_get_attention(
    heads: Optional[Union[int, Iterable[int], str]] = None,
    output_idx: Optional[int] = None,
    ) -> Callable:
    """Creates a hooking function to get the attention patterns of a given layer.

    :param heads: The heads for which to save the attention, defaults to None
    :type heads: Optional[Union[int, Iterable[int], str]], optional
    :param output_idx: If the attention module returns a tuple, use this argument to index it, defaults to None
    :type output_idx: Optional[int], optional
    :return: func, hooking function that saves attention of the specified heads
    :rtype: Callable
    """
    
    # convert head string to slice
    if heads is None:
        heads = ":"
    if isinstance(heads, str):
        heads = util.create_slice_from_str(heads)

    def func(save_ctx, input, output):
        if output_idx is None:
            save_ctx['attn'] = output[:,heads,...].detach().cpu()
        else:
            save_ctx['attn'] = output[output_idx][:,heads,...].detach().cpu()
    
    return func

def create_attention_hook(
    layer: int, 
    key: str, 
    output_idx: Optional[int] = None,
    attn_name: Optional[str] = 'attn',
    layer_key_prefix: Optional[str] = None,
    heads: Optional[Union[int, Iterable[int], str]] = None
) -> Hook:
    """Creates a hook which saves the attention patterns of a given layer.

    :param layer: The layer to hook.
    :type layer: int
    :param key: The key to use for saving the attention patterns.
    :type key: str
    :param output_idx: If the module output is a tuple, index it with this. GPT like models need this to be equal to 2, defaults to None
    :type output_idx: Optional[int], optional
    :param attn_name: The name of the attention module in the transformer, defaults to 'attn'
    :type attn_name: Optional[str], optional
    :param layer_key_prefix: The prefix in the model structure before the layer idx, e.g. 'transformer->h', defaults to None
    :type layer_key_prefix: Optional[str], optional
    :param heads: Which heads to save the attention pattern for. Can be int, tuple of ints or string like '1:3', defaults to None
    :type heads: Optional[Union[int, Iterable[int], str]], optional
    :return: Hook which saves the attention patterns
    :rtype: Hook
    """
    if layer_key_prefix is None:
        layer_key_prefix = ""
        
    func = transformers_get_attention(heads, output_idx)
    return Hook(f'{layer_key_prefix}{layer}->{attn_name}', func, key)


def create_logit_hook(
    layer:int, 
    model: HookedModel, 
    unembedding_key: str,
    layer_key_prefix: Optional[str] = None,
    target: Optional[Union[int, List[int]]] = None, 
    position: Optional[Union[int, List[int]]] = None,
    key: Optional[str] = None,
    split_heads: Optional[bool] = False,
    num_heads: Optional[int] = None,
) -> Hook:
    """Create a hook that saves the logits of a layer's output.
    Outputs are saved to save_ctx[key]['logits'].
    
    :param layer: The number of the layer
    :type layer: int
    :param model: The model.
    :type model: HookedModel
    :param unembedding_key: The key/name of the embedding matrix, e.g. 'lm_head' for causal LM models
    :type unembedding_key: str
    :param layer_key_prefix: The prefix of the key of the layer, e.g. 'transformer->h' for GPT like models
    :type layer_key_prefix: str
    :param target: The target token(s) to extract logits for. Defaults to all tokens.
    :type target: Union[int, List[int]]
    :param position: The position for which to extract logits for. Defaults to all positions.
    :type position: Union[int, List[int]]
    :param key: The key of the hook. Defaults to {layer}_logits.
    :type key: str
    :param split_heads: Whether to split the heads. Defaults to False.
    :type split_heads: bool
    :param num_heads: The number of heads to split. Defaults to None.
    :type num_heads: int
    :return: The hook.
    :rtype: Hook
    """
    if layer_key_prefix is None:
        layer_key_prefix = ""
        
    # generate slice
    if target is None:
        target = ":"
    else:
        if isinstance(target, int):
            target = str(target)
        else:
            target = "[" + ",".join(str(t) for t in target) + "]"
    if position is None:
        position = ":"
    else:
        if isinstance(position, int):
            position = str(position)
        else:
            position = "[" + ",".join(str(p) for p in position) + "]"
    position_slice = util.create_slice_from_str(f":,{position},:")
    target_slice = util.create_slice_from_str(f"{target},:")
    
    # load the relevant part of the vocab matrix
    vocab_matrix = model.layers[unembedding_key].weight[target_slice].T
        
    # split vocab matrix
    if split_heads:
        if num_heads is None:
            raise ValueError("num_heads must be specified if split_heads is True")
        else:
            vocab_matrix = einops.rearrange(vocab_matrix, '(num_heads head_dim) vocab_size -> num_heads head_dim vocab_size', num_heads=num_heads)
    
    def inner(save_ctx, input, output):
        if split_heads:
            einsum_in = einops.rearrange(output[0][position_slice], 'batch seq_len (num_heads head_dim) -> batch num_heads seq_len head_dim', num_heads=num_heads)
            einsum_out = torch.einsum('bcij,cjk->bcik', einsum_in, vocab_matrix)
        else:
            einsum_in = output[0][position_slice]
            einsum_out = torch.einsum('bij,jk->bik', einsum_in, vocab_matrix)
            
        save_ctx['logits'] = einsum_out.detach().cpu()
    
    # write key
    if key is None:
        key = str(layer) + '_logits'
    
    # create hook
    hook = Hook(f'{layer_key_prefix}{layer}', inner, key)
    
    return hook

def gpt_attn_wrapper(
    func: Callable, 
    save_ctx: Dict, 
    c_proj: torch.Tensor, 
    vocab_embedding: torch.Tensor, 
    target_ids: torch.Tensor,
    batch_size: Optional[int] = None,
) -> Tuple[Callable, Callable]:
    """Wraps around the [AttentionBlock]._attn function to save the individual heads' logits.
    This is necessary because the individual heads' logits are not available on a module level and thus not accessible via a hook.

    :param func: original _attn function
    :type func: Callable
    :param save_ctx: context to which the logits will be saved
    :type save_ctx: Dict
    :param c_proj: projection matrix, this is W_O in Anthropic's terminology
    :type c_proj: torch.Tensor
    :param vocab_matrix: vocabulary/embedding matrix, this is W_V in Anthropic's terminology
    :type vocab_matrix: torch.Tensor
    :param target_ids: indices of the target tokens for which the logits are computed
    :type target_ids: torch.Tensor
    :param batch_size: batch size to reduce memory footprint, defaults to None
    :type batch_size: Optional[int]
    :return: inner, func, the wrapped function and the original function
    :rtype: Tuple[Callable, Callable]
    """
    # TODO Find a smarter/more efficient way of implementing this function
    # TODO clean up this function
    def inner(query, key, value, *args, **kwargs):
        nonlocal c_proj
        nonlocal target_ids
        nonlocal vocab_embedding
        nonlocal batch_size
        attn_output, attn_weights = func(query, key, value, *args, **kwargs)
        if batch_size is None:
            batch_size = attn_output.shape[0]
        with torch.no_grad():
            temp = attn_weights[...,None] * value[:,:,None]
            if len(c_proj.shape) == 2:
                c_proj = einops.rearrange(c_proj, '(head_dim num_heads) out_dim -> head_dim num_heads out_dim', num_heads=attn_output.shape[1])
            c_proj = einops.rearrange(c_proj, 'h n o -> n h o')
            temp = temp[0,:,:-1,:-1] # could this be done earlier?
            new_temp = []
            for head in tqdm(range(temp.shape[0])):
                new_temp.append([])
                for i in range(math.ceil(temp.shape[1] / batch_size)):
                    out = temp[head, i*batch_size:(i+1)*batch_size] @ c_proj[head]
                    out = out @ vocab_embedding # compute logits
                    new_temp[-1].append(out)
            
                # center logits
                new_temp[-1] = torch.cat(new_temp[-1])
                new_temp[-1] -= torch.mean(new_temp[-1], dim=-1, keepdim=True)
                # select targets
                new_temp[-1] = new_temp[-1][...,torch.arange(len(target_ids)), :, target_ids]#.to('cpu')

            new_temp = torch.stack(new_temp, dim=0) # stack heads
            max_pos_value = torch.amax(new_temp).item()
            max_neg_value = torch.amax(-new_temp).item()
            
            save_ctx['logits'] = {
                'pos': (new_temp/max_pos_value).clamp(min=0, max=1).detach().cpu(),
                'neg': (-new_temp/max_neg_value).clamp(min=0, max=1).detach().cpu(),
            }         
        return attn_output, attn_weights
    return inner, func

#TODO update docs here
def additive_output_noise(
    indices: str, 
    mean: Optional[float] = 0, 
    std: Optional[float] = 0.1
) -> Callable:
    slice_ = util.create_slice_from_str(indices)
    def func(save_ctx, input, output):
        noise = mean + std * torch.randn_like(output[slice_])
        output[slice_] += noise
        return output
    return func

def hidden_patch_hook_fn(
    position: int, 
    replacement_tensor: torch.Tensor,
) -> Callable:
    indices = "...," + str(position) + len(replacement_tensor.shape) * ",:"
    inner = replace_activation(indices, replacement_tensor)    
    def func(save_ctx, input, output):
        output[0][...] = inner(save_ctx, input, output[0])
        return output



from collections import OrderedDict
from dataclasses import dataclass
from inspect import signature
from typing import List, Callable

import torch

from . import util

class Hook:
    layer_name: str
    func: Callable
    key: str
    def __init__(self, layer_name: str, func: Callable, key: str):
        # check that func takes three arguments
        if len(signature(func).parameters) != 3:
            raise TypeError(f'Hook function {func.__name__} should have three arguments, but has {len(signature(func).parameters)}.')

        self.layer_name = layer_name
        self.func = func
        self.key = key

class HookedModel(torch.nn.Module):
    def __init__(self, model):
        """Wrapper around a module that allows forward passes with hooks and a context object.

        :param model: Model to be hooked
        :type model: nn.Module
        :raises TypeError: Incorrect model type
        """
        super().__init__()

        # check inputs
        if not isinstance(model, torch.nn.Module):
            raise TypeError(f"model should be type torch.nn.Module but is {type(model)}")
        
        self.model = model
        
        # initialize hooks
        self.init_refs()

        # init context for accessing hook output
        self.save_ctx = dict()

    def init_refs(self):
        """Creates references for every layer in a model."""

        self.layers = OrderedDict()

        # recursive naming function
        def name_layers(net, prefix=[]):
            if hasattr(net, "_modules"):
                for name, layer in net._modules.items():
                    if layer is None:
                        # e.g. GoogLeNet's aux1 and aux2 layers
                        continue
                    self.layers["->".join(prefix + [name])] = layer
                    name_layers(layer, prefix=prefix + [name])
            else:
                raise ValueError('net has not _modules attribute! Check if your model is properly instantiated..')

        name_layers(self.model)


    def forward(
        self, 
        input_ids: torch.Tensor, 
        hooks: List[Hook],
        *args,
        **kwargs,
    ):
        """Wrapper around the default forward pass that temporarily registers hooks, executes the forward pass and then closes hooks again.
        """
        # register hooks
        registered_hooks = []
        for hook in hooks:
            layer = self.layers.get(hook.layer_name, None)
            if layer is None:
                raise ValueError(f'Layer {hook.layer_name} was not found during hook registration! Here is the whole model for reference:\n {self.__repr__}')
            self.save_ctx[hook.key] = dict() # create sub-context for each hook to write to
            registered_hooks.append(layer.register_forward_hook(self._hook_wrapper(hook.func, hook.key)))

        # forward
        output = self.model(input_ids, *args, **kwargs) #TODO generalize to non-HF models which would not have an input_ids kwarg

        # remove hooks
        for hook in registered_hooks:
            hook.remove()
            
        return output

    def _hook_wrapper(self, func, hook_key):
        """Wrapper to comply with PyTorch's hooking API while enabling saving to context.

        :param func: [description]
        :type func: [type]
        :param hook_key: [description]
        :type hook_key: [type]
        :return: [description]
        :rtype: [type]
        """     
        return lambda model, input, output: func(save_ctx=self.save_ctx[hook_key], input=input[0], output=output)

    def get_ctx_keys(self):
        return list(self.save_ctx.keys())

    def __repr__(self):
        return self.model.__repr__()
    
    @property
    def device(self):
        return next(self.model.parameters()).device


import time
import json
import streamlit as st

from unseal.visuals.streamlit_interfaces import utils
from unseal.visuals.streamlit_interfaces import interface_setup as setup
from unseal.visuals.streamlit_interfaces.commons import SESSION_STATE_VARIABLES

# perform startup tasks
setup.startup(SESSION_STATE_VARIABLES, './registered_models.json')

# create sidebar
with st.sidebar:
    setup.create_sidebar()
    
    sample = st.checkbox('Enable sampling', value=False, key='sample')
    if sample:
        setup.create_sample_sliders()
        setup.on_sampling_config_change()
    
    if "storage" not in st.session_state:
        st.session_state["storage"] = ["", ""]
        
    # input 1
    placeholder1 = st.empty()
    placeholder1.text_area(label='Input 1', on_change=utils.on_text_change, key='input_text_1', value=st.session_state.storage[0], kwargs=dict(col_idx=0, text_key='input_text_1'))
    if sample:
        st.button(label="Sample", on_click=utils.sample_text, kwargs=dict(model=st.session_state.model, col_idx=0, key="input_text_1"), key="sample_text_1")
    
    # input 2
    placeholder2 = st.empty()
    placeholder2.text_area(label='Input 2', on_change=utils.on_text_change, key='input_text_2', value=st.session_state.storage[1], kwargs=dict(col_idx=1, text_key='input_text_2'))
    if sample:
        st.button(label="Sample", on_click=utils.sample_text, kwargs=dict(model=st.session_state.model, col_idx=1, key="input_text_2"), key="sample_text_2")
    
    # sometimes need to force a re-render
    st.button('Show Attention', on_click=utils.text_change, kwargs=dict(col_idx=[0,1]))
    
    # download button
    f =  json.encoder.JSONEncoder().encode(st.session_state.visualization)
    st.download_button(
        label='Download Visualization', 
        data=f, 
        file_name=f'{st.session_state.model_name}_{time.strftime("%Y%m%d_%H%M%S", time.localtime())}.json', 
        mime='application/json', 
        help='Download the visualizations as a json of html files.', 
        key='download_button'
    )

# show the html visualization
if st.session_state.model is not None:
    cols = st.columns(2)
    for col_idx, col in enumerate(cols):
        if f"col_{col_idx}" in st.session_state.visualization:
            with col:
                for layer in range(st.session_state.num_layers):
                    if f"layer_{layer}" in st.session_state.visualization[f"col_{col_idx}"]:
                        with st.expander(f'Layer {layer}'):
                            st.components.v1.html(st.session_state.visualization[f"col_{col_idx}"][f"layer_{layer}"], height=600, scrolling=True)
        else:
            st.session_state.visualization[f"col_{col_idx}"] = dict()



def layer_change():
    utils.text_change(0)
    
import json
import time

import streamlit as st

from unseal.visuals.streamlit_interfaces import utils
from unseal.visuals.streamlit_interfaces import interface_setup as setup
from unseal.visuals.streamlit_interfaces.commons import SESSION_STATE_VARIABLES

# perform startup tasks
setup.startup(SESSION_STATE_VARIABLES, './registered_models.json')

# create sidebar
with st.sidebar:
    setup.create_sidebar()
    
    sample = st.checkbox('Enable sampling', value=False, key='sample')
    if sample:
        setup.create_sample_sliders()
        setup.on_sampling_config_change()
    
    if "storage" not in st.session_state:
        st.session_state["storage"] = [""]
    
    # select layer
    if st.session_state.num_layers is None:
        options = list(['Select a model!'])
    else:
        options = list(range(st.session_state.num_layers))
    st.selectbox('Layer', options=options, key='layer', on_change=layer_change, index=0)
    
    # input 1
    placeholder = st.empty()
    placeholder.text_area(label='Input', on_change=utils.on_text_change, key='input_text', value=st.session_state.storage[0], kwargs=dict(col_idx=0, text_key='input_text'))
    if sample:
        st.button(label="Sample", on_click=utils.sample_text, kwargs=dict(model=st.session_state.model, col_idx=0, key="input_text"), key="sample_text")
    
    # sometimes need to force a re-render
    st.button('Show Attention', on_click=utils.text_change, kwargs=dict(col_idx=0))
    
    f =  json.encoder.JSONEncoder().encode(st.session_state.visualization)
    st.download_button(
        label='Download Visualization', 
        data=f, 
        file_name=f'{st.session_state.model_name}_{time.strftime("%Y%m%d_%H%M%S", time.localtime())}.json', 
        mime='application/json', 
        help='Download the visualizations as a json of html files.', 
        key='download_button'
    )

# show the html visualization
if st.session_state.model is not None:
    cols = st.columns(1)
    for col_idx, col in enumerate(cols):
        if f"col_{col_idx}" in st.session_state.visualization:
            with col:
                if f"layer_{st.session_state.layer}" in st.session_state.visualization[f"col_{col_idx}"]:
                    with st.expander(f'Layer {st.session_state.layer}'):
                        st.components.v1.html(st.session_state.visualization[f"col_{col_idx}"][f"layer_{st.session_state.layer}"], height=600)
        else:
            st.session_state.visualization[f"col_{col_idx}"] = dict()


from setuptools import setup, find_packages
import os, sys
sys.path.insert(0, os.path.abspath("."))
from unseal import __version__

version = __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='Unseal',
    version=version,
    packages=find_packages(exclude=[]),
    python_requires='>=3.6.0',
    install_requires=[
        'torch',
        'einops>=0.3.2',
        'numpy',
        'transformers',
        'tqdm',
        'matplotlib',
        'streamlit',
    ],
    # entry_points={
    #     'console_scripts': [
    #         '"unseal compare" = unseal.commands.interfaces.compare_two_inputs:main',
    #     ]
    # },
    description=(
        "Unseal: "
        "A collection of infrastructure and tools for research in "
        "transformer interpretability."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="The Unseal Team",
    author_email="tlieberum@outlook.de",
    url="https://github.com/TomFrederik/unseal/",
    license="Apache License 2.0",
    keywords=[
        "pytorch",
        "tensor",
        "machine learning",
        "neural networks",
        "interpretability",
        "transformers",
    ],
)



from typing import Union, List

import streamlit as st
from transformers.models.gpt_neo.modeling_gpt_neo import GPTNeoPreTrainedModel

from ..utils import compute_attn_logits

def sample_text(model, col_idx, key):
    text = st.session_state[key]
    if st.session_state.prefix_prompt is not None and len(st.session_state.prefix_prompt) > 0:
        text = st.session_state.prefix_prompt + '\n' + text
    model_inputs = st.session_state.tokenizer.encode(text, return_tensors='pt').to(st.session_state.device)
    output = model.model.generate(model_inputs, **st.session_state.sample_kwargs, min_length=0, output_attentions=True)
    output_text = st.session_state.tokenizer.decode(output[0], skip_special_tokens=True)
    if st.session_state.prefix_prompt is not None and len(st.session_state.prefix_prompt) > 0:
        output_text = output_text.lstrip(st.session_state.prefix_prompt + '\n')
    
    st.session_state["storage"][col_idx] = output_text
    text_change(col_idx=col_idx)


def on_text_change(col_idx: Union[int, List[int]], text_key):
    if isinstance(col_idx, list):
        for idx in col_idx:
            on_text_change(idx, text_key)
    else:    
        st.session_state["storage"][col_idx] = st.session_state[text_key]
        text_change(col_idx)

def get_attn_logits_args():
    # get args for compute_attn_logits
    if st.session_state.model_name in st.session_state.registered_model_names:
        attn_name = st.session_state.config['attn_name']
        output_idx = st.session_state.config['output_idx']
        layer_key_prefix = st.session_state.config['layer_key_prefix']
        out_proj_name = st.session_state.config['out_proj_name']
        attn_suffix = st.session_state.config['attn_suffix']
        unembedding_key = st.session_state.config['unembedding_key']
    elif isinstance(st.session_state.model.model, GPTNeoPreTrainedModel):
        attn_name = 'attn'
        output_idx = 2
        layer_key_prefix = 'transformer->h'
        out_proj_name = 'out_proj'
        attn_suffix = 'attention'
        unembedding_key = 'lm_head'
    else:
        attn_name = 'attn'
        output_idx = 2
        layer_key_prefix = 'transformer->h'
        out_proj_name = 'c_proj'
        attn_suffix = None
        unembedding_key = 'lm_head'
    return attn_name, output_idx, layer_key_prefix, out_proj_name, attn_suffix, unembedding_key

def text_change(col_idx: Union[int, List[int]]):
    if isinstance(col_idx, list):
        for idx in col_idx:
            text_change(idx)
        return
    
    text = st.session_state["storage"][col_idx]
    if st.session_state.prefix_prompt is not None and len(st.session_state.prefix_prompt) > 0:
        text = st.session_state.prefix_prompt + '\n' + text

    if text is None or len(text) == 0:
        return
    
    attn_name, output_idx, layer_key_prefix, out_proj_name, attn_suffix, unembedding_key = get_attn_logits_args()
    
    if 'layer' in st.session_state:
        layer = st.session_state['layer']
    else:
        layer = None
    
    compute_attn_logits(
        st.session_state.model, 
        st.session_state.model_name, 
        st.session_state.tokenizer,
        st.session_state.num_layers,
        text, 
        st.session_state.visualization[f'col_{col_idx}'],
        attn_name = attn_name,
        output_idx = output_idx,
        layer_key_prefix = layer_key_prefix,
        out_proj_name = out_proj_name,
        attn_suffix = attn_suffix,
        unembedding_key = unembedding_key,
        layer_id = layer,
    )


# utility functions for interacting with huggingface's transformers library
import logging
import os
from typing import Optional, Tuple

from transformers import AutoTokenizer, AutoConfig, AutoModelForCausalLM
from transformers.file_utils import RepositoryNotFoundError

from .hooks.commons import HookedModel

def load_from_pretrained(
    model_name: str, 
    model_dir: Optional[str] = None, 
    load_model: Optional[bool] = True, 
    load_tokenizer: Optional[bool] = True, 
    load_config: Optional[bool] = True,
    low_cpu_mem_usage: Optional[bool] = False,
) -> Tuple[AutoModelForCausalLM, AutoTokenizer, AutoConfig]:
    """Load a pretrained model from huggingface's transformer library

    :param model_name: Name of the model, e.g. `gpt2` or `gpt-neo`.
    :type model_name: str
    :param model_dir: Directory in which to look for the model, e.g. `EleutherAI`, defaults to None
    :type model_dir: Optional[str], optional
    :param load_model: Whether to load the model itself, defaults to True
    :type load_model: Optional[bool], optional
    :param load_tokenizer: Whether to load the tokenizer, defaults to True
    :type load_tokenizer: Optional[bool], optional
    :param load_config: Whether to load the config file, defaults to True
    :type load_config: Optional[bool], optional
    :param low_cpu_mem_usage: Whether to use low-memory mode, experimental feature of HF, defaults to False
    :type low_cpu_mem_usage: bool, optional
    :return: model, tokenizer, config. Returns None values for those elements which were not loaded.
    :rtype: Tuple[AutoModelForCausalLM, AutoTokenizer, AutoConfig]
    """
    if model_dir is None:
        try:
            logging.info(f'Loading model {model_name}')

            model = AutoModelForCausalLM.from_pretrained(model_name, low_cpu_mem_usage=low_cpu_mem_usage) if load_model else None        
            tokenizer = AutoTokenizer.from_pretrained(model_name) if load_tokenizer else None
            config = AutoConfig.from_pretrained(model_name) if load_config else None

        except (RepositoryNotFoundError, OSError) as error:
            logging.warning("Couldn't find model in default folder. Trying EleutherAI/...")

            model = AutoModelForCausalLM.from_pretrained(f'EleutherAI/{model_name}', low_cpu_mem_usage=low_cpu_mem_usage) if load_model else None
            tokenizer = AutoTokenizer.from_pretrained(f'EleutherAI/{model_name}') if load_tokenizer else None
            config = AutoConfig.from_pretrained(f'EleutherAI/{model_name}') if load_config else None

    else:
        model = AutoModelForCausalLM.from_pretrained(os.path.join(model_dir, model_name), low_cpu_mem_usage=low_cpu_mem_usage) if load_model else None        
        tokenizer = AutoTokenizer.from_pretrained(os.path.join(model_dir, model_name)) if load_tokenizer else None
        config = AutoConfig.from_pretrained(os.path.join(model_dir, model_name)) if load_config else None

    return model, tokenizer, config

def get_num_layers(model: HookedModel, layer_key_prefix: Optional[str] = None) -> int:
    """Get the number of layers in a model

    :param model: The model to get the number of layers from
    :type model: HookedModel
    :param layer_key_prefix: The prefix to use for the layer keys, defaults to None
    :type layer_key_prefix: Optional[str], optional
    :return: The number of layers in the model
    :rtype: int
    """
    if layer_key_prefix is None:
        layer_key_prefix = ""
    
    return len(model.layers[f"{layer_key_prefix}"])


import json
from io import StringIO

import streamlit as st

from unseal.visuals.streamlit_interfaces import interface_setup as setup
from unseal.visuals import utils

def on_file_upload():
    if st.session_state.uploaded_file is None:
        return
    
    data = json.loads(StringIO(st.session_state.uploaded_file.getvalue().decode('utf-8')).read())
    
    st.session_state.visualization = data
    
    for layer in range(len(data)):
        html_str = st.session_state.visualization[f'layer_{layer}']
        with st.expander(f'Layer {layer}'):
            st.components.v1.html(html_str, height=600, scrolling=True)

# set page config to wide layout
st.set_page_config(layout='wide')

# create sidebar
with st.sidebar:
    st.file_uploader(
        label='Upload Visualization', 
        accept_multiple_files=False,
        on_change=on_file_upload,
        help='Upload the visualizations as an json of html files.', 
        key='uploaded_file'
    )


def load():
    with open("./visualizations/gpt2-xl.json", "r") as fp:
        data = json.load(fp)

    st.session_state.visualization = data
    for layer in range(len(data['col_0'])):
        html_str = st.session_state.visualization['col_0'][f'layer_{layer}']
        with st.expander(f'Layer {layer}'):
            st.components.v1.html(html_str, height=1000, width=800, scrolling=True)

load()




import json
import gc
from typing import Optional, Callable, Dict

import einops
import pysvelte as ps
import torch
from transformers import AutoTokenizer

from ..hooks import HookedModel
from ..hooks.common_hooks import create_attention_hook, gpt_attn_wrapper

def compute_attn_logits(
    model: HookedModel,
    model_name: str, 
    tokenizer: AutoTokenizer, 
    num_layers: int, 
    text: str, 
    html_storage: Dict, 
    save_path: Optional[str] = None,
    attn_name: Optional[str] = 'attn',
    output_idx: Optional[int] = 2,
    layer_key_prefix: Optional[str] = None,
    out_proj_name: Optional[str] = 'out_proj',
    attn_suffix: Optional[str] = None,
    unembedding_key: Optional[str] = 'lm_head',
    layer_id: Optional[int] = None,
    batch_size: Optional[int] = None,
):
    # parse inputs
    if save_path is None:
        save_path = f"{model_name}.json" 
    if layer_key_prefix is None:
        layer_key_prefix = ""
    else:
        layer_key_prefix += "->"
    if attn_suffix is None or attn_suffix == "":
        attn_suffix = ""
    if layer_id is None:
        layer_list = list(range(num_layers))
    else:
        layer_list = [layer_id]
    
    # tokenize without tokenization artifact -> needed for visualization
    tokenized_text = tokenizer.tokenize(text)
    tokenized_text = list(map(tokenizer.convert_tokens_to_string, map(lambda x: [x], tokenized_text))) 
    
    # encode text
    model_input = tokenizer.encode(text, return_tensors='pt').to(model.device)
    target_ids = tokenizer.encode(text)[1:]
    
    # compute attention pattern
    attn_hooks = [create_attention_hook(layer, f'attn_layer_{layer}', output_idx, attn_name, layer_key_prefix) for layer in layer_list]
    model.forward(model_input, hooks=attn_hooks, output_attentions=True)
    
    # compute logits
    for layer in layer_list:

        # wrap the _attn function to create logit attribution
        model.save_ctx[f'logit_layer_{layer}'] = dict()
        old_fn = wrap_gpt_attn(model, layer, target_ids, unembedding_key, attn_name, attn_suffix, layer_key_prefix, out_proj_name, batch_size)

        # forward pass
        model.forward(model_input, hooks=[])
    
        # parse attentions for this layer
        attention = model.save_ctx[f"attn_layer_{layer}"]['attn'][0]
        attention = einops.rearrange(attention, 'h n1 n2 -> n1 n2 h')

        # parse logits
        if model_input.shape[1] > 1: # otherwise we don't have any logit attribution
            logits = model.save_ctx[f'logit_layer_{layer}']['logits']
            pos_logits = logits['pos']
            neg_logits = logits['neg']
            pos_logits = pad_logits(pos_logits)
            neg_logits = pad_logits(neg_logits)
            
            pos_logits = einops.rearrange(pos_logits, 'h n1 n2 -> n1 n2 h')
            neg_logits = einops.rearrange(neg_logits, 'h n1 n2 -> n1 n2 h')
        else:
            pos_logits = torch.zeros((attention.shape[0], attention.shape[1], attention.shape[2]))
            neg_logits = torch.zeros((attention.shape[0], attention.shape[1], attention.shape[2]))
        
        # compute and display the html object
        html_object = ps.AttentionLogits(tokens=tokenized_text, attention=attention, pos_logits=pos_logits, neg_logits=neg_logits, head_labels=[f'{layer}:{j}' for j in range(attention.shape[-1])])
        html_object = html_object.update_meta(suppress_title=True)
        html_str = html_object.html_page_str()

        # save html string
        html_storage[f'layer_{layer}'] = html_str
        
        # reset _attn function
        reset_attn_fn(model, layer, old_fn, attn_name, attn_suffix, layer_key_prefix)
        
        # save progress so far
        with open(save_path, "w") as f: 
            json.dump(html_storage, f)

        # garbage collection
        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            
    return html_storage

def pad_logits(logits):
    logits = torch.cat([torch.zeros_like(logits[:,0,None]), logits], dim=1)
    logits = torch.cat([logits, torch.zeros_like(logits[:,:,0,None])], dim=2)
    return logits
    
def wrap_gpt_attn(
    model: HookedModel, 
    layer: int, 
    target_ids: Callable,
    unembedding_key: str,
    attn_name: Optional[str] = 'attn',
    attn_suffix: Optional[str] = None,
    layer_key_prefix: Optional[str] = None,
    out_proj_name: Optional[str] = 'out_proj',
    batch_size: Optional[int] = None,
) -> Callable:
    # parse inputs
    if layer_key_prefix is None:
        layer_key_prefix = ""
    if attn_suffix is None:
        attn_suffix = ""

    attn_name = f"{layer_key_prefix}{layer}->{attn_name}{attn_suffix}"
    out_proj_name = attn_name + f"->{out_proj_name}"
    
    model.layers[attn_name]._attn, old_fn = gpt_attn_wrapper(
        model.layers[attn_name]._attn,
        model.save_ctx[f'logit_layer_{layer}'], 
        model.layers[out_proj_name].weight,
        model.layers[unembedding_key].weight.T,
        target_ids,
        batch_size,
    )
    
    return old_fn
        

def reset_attn_fn(
    model: HookedModel, 
    layer: int, 
    old_fn: Callable,
    attn_name: Optional[str] = 'attn',
    attn_suffix: Optional[str] = None,
    layer_key_prefix: Optional[str] = None,
) -> None:
    # parse inputs
    if layer_key_prefix is None:
        layer_key_prefix = ""
    if attn_suffix is None:
        attn_suffix = ""
    
    # reset _attn function to old_fn
    del model.layers[f"{layer_key_prefix}{layer}->{attn_name}{attn_suffix}"]._attn
    model.layers[f"{layer_key_prefix}{layer}->{attn_name}{attn_suffix}"]._attn = old_fn